import json
import os
from argo.workflows.sdk import Workflow, template
from asgiref.sync import async_to_sync
# This is just a K8s V1Container.
from argo.workflows.sdk.templates import V1Container
from eliot import start_action
from kubernetes.client import V1ResourceRequirements, V1SecurityContext
from kubernetes.client.rest import ApiException
from kubernetes import client
from .. import LoggableChild
from ..utils import list_digest, str_true


class LSSTWorkflowManager(LoggableChild):
    '''This is an LSST Manager Class containing LSST (er, Rubin Observatory)-
    specific logic regarding management of Argo Workflows.
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cmd_vol = None
        self.cmd_mt = None
        self.workflow = None
        self.cfg_map = None
        self.wf_input = None

    def define_configmap(self, data):
        '''This returns a k8s configmap using the data from the new-workflow
        POST.
        '''
        with start_action(action_type="define_configmap"):
            ni_cmd = data['command']
            idkey = list_digest(ni_cmd)
            cm_name = 'command.{}.json'.format(idkey)
            k8s_vol = client.V1Volume(
                name="noninteractive-command",
                config_map=client.V1ConfigMapVolumeSource(
                    name=cm_name
                ),
            )
            k8s_mt = client.V1VolumeMount(
                name="noninteractive-command",
                mount_path=("/opt/lsst/software/jupyterlab/" +
                            "noninteractive/command/"),
                read_only=True
            )
            self.cmd_vol = k8s_vol
            self.cmd_mt = k8s_mt
            # Now the configmap
            cm_data = {}
            cm_data.update(data)
            del cm_data['image']
            del cm_data['size']
            jd = json.dumps(data, sort_keys=True, indent=4)
            k8s_configmap = client.V1ConfigMap(
                metadata=client.V1ObjectMeta(name=cm_name),
                data={'command.json': json.dumps(data)})
            self.log.debug("Created configmap '{}': {}".format(cm_name, jd))
            self.cfg_map = k8s_configmap

    def define_workflow(self, data):
        '''This is basically our equivalent of get_pod_manifest().
        It creates a dict which we will pass to the workflow template
        engine, which will allow it to create an appropriate workflow.
        '''
        with start_action(action_type="define_workflow"):
            wf_input = {}
            # FIXME Right now we can assume data is of type 'cmd'; we need
            # a little tweaking for 'nb' in that the command will be fixed
            # and the execution parameters will differ.
            cfg = self.parent.config
            em = self.parent.env_mgr
            vm = self.parent.volume_mgr
            am = self.parent.auth_mgr
            om = self.parent.optionsform_mgr
            user = self.parent.parent.authenticator.user
            # Definitely not spawner in this case--spawner is mocked out.
            em_env = em.get_env()
            size_map = om.sizemap.get(data['size'])
            if not size_map:
                # Use default size
                cpu = cfg.tiny_cpu_max * (2 ** cfg.size_index)
                mem = cfg.mb_per_cpu * cpu
                cpu_guar = cpu / (float(cfg.lab_size_range))
                mem_guar = cfg.mb_per_cpu * cpu_guar
                size_map = {
                    "cpu": str(cpu),
                    "mem": "{}M".format(mem),
                    "cpu_guar": str(cpu_guar),
                    "mem_guar": "{}M".format(mem_guar)
                }
            self.log.debug(
                "Size '{}' resolves to '{}'".format(data['size'], size_map))
            ml = size_map['mem']
            cl = size_map['cpu']
            mg = size_map['mem_guar']
            cg = size_map['cpu_guar']
            # If we actually have this in the environment, use it
            # (Dask requires due to our configuration).  If we do not,
            # then make up something plausible, but it really doesn't
            # matter if we get it right, because it's not like we're
            # using the Dask proxy dashboard from inside a Workflow anyway.
            synth_jsp = '/nb/user/{}'.format(user.escaped_name)
            jsp = os.getenv('JUPYTERHUB_SERVER_PREFIX', synth_jsp)
            wf_input['mem_limit'] = ml
            wf_input['mem_guar'] = mg
            wf_input['cpu_limit'] = str(cl)
            wf_input['cpu_guar'] = str(cg)
            wf_input['image'] = data['image']
            wf_input['enable_multus'] = cfg.enable_multus
            env = {}
            vols = []
            vmts = []
            env.update(em_env)
            vols.extend(vm.k8s_volumes)
            vmts.extend(vm.k8s_vol_mts)
            self.define_configmap(data)
            vols.append(self.cmd_vol)
            vmts.append(self.cmd_mt)
            env['DASK_VOLUME_B64'] = vm.get_dask_volume_b64()
            cname = "wf-{}-{}-{}".format(
                user.escaped_name,
                data['image'].split(':')[-1].replace('_', '-'),
                data['command'][0].split('/')[-1].replace('_', '-'))
            wf_input['name'] = cname
            env['MEM_LIMIT'] = ml
            env['MEM_GUARANTEE'] = mg
            env['CPU_LIMIT'] = str(cl)
            env['CPU_GUARANTEE'] = str(cg)
            env['JUPYTERHUB_USER'] = user.escaped_name
            env['NONINTERACTIVE'] = "TRUE"
            ast = async_to_sync(user.get_auth_state)()
            env['EXTERNAL_UID'] = str(ast['uid'])
            env['EXTERNAL_GROUPS'] = am.get_group_string()
            env['JUPYTERHUB_SERVER_PREFIX'] = jsp
            env['DEBUG'] = str_true(cfg.debug)
            # Get token, if we have one.  We want the one from the launching
            #  lab container, if it exists.
            # Try file first

            token = self._get_access_token()
            if token:
                env['ACCESS_TOKEN'] = token
            e_l = self._d2l(env)
            wf_input['env'] = e_l
            wf_input['username'] = user.escaped_name
            # Volumes and mounts aren't JSON-serializable...
            wf_input['vols'] = '{}'.format(vols)
            wf_input['vmts'] = '{}'.format(vmts)
            self.wf_input = wf_input
            self.log.debug("Input to Workflow Manipulator: {}".format(
                json.dumps(wf_input, indent=4, sort_keys=True)))
            # ...now put the real values back
            wf_input['vols'] = vols
            wf_input['vmts'] = vmts
            wf_input['command'] = data['command']
            self.wf_input = wf_input
            wf = LSSTWorkflow(parms=wf_input)
            self.log.debug("Workflow: {}".format(wf))
            self.workflow = wf

    def _d2l(self, in_d):
        with start_action(action_type="_d2l"):
            ll = []
            for k in in_d:
                ll.append({"name": k,
                           "value": in_d[k]})
            return ll

    def _get_access_token(self):
        tok = None
        hdir = os.environ.get('HOME', None)
        if hdir:
            tokfile = hdir + "/.access_token"
            try:
                with open(tokfile, 'r') as f:
                    tok = f.read().replace('\n', '')
            except Exception as exc:
                self.log.warning(
                    "Could not read token from '{}': {}'".format(tokfile, exc))
        if not tok:
            tok = os.environ.get('ACCESS_TOKEN', None)
        return tok

    def list_workflows(self):
        with start_action(action_type="list_workflows"):
            namespace = self.parent.namespace_mgr.namespace
            api = self.parent.wf_api
            self.log.debug(
                "Listing workflows in namespace '{}'".format(namespace))
            nl = self.parent.api.list_namespace(timeout_seconds=1)
            found = False
            for ns in nl.items:
                nsname = ns.metadata.name
                if nsname == namespace:
                    self.log.debug("Namespace {} found.".format(namespace))
                    found = True
                    break
            if not found:
                self.log.debug("No namespace {} found.".format(namespace))
                wfs = None
            else:
                wfs = api.list_namespaced_workflows(namespace=namespace)
            return wfs

    def create_workflow(self):
        with start_action(action_type="create_workflows"):
            workflow = self.workflow
            namespace = self.parent.namespace_mgr.namespace
            api = self.parent.wf_api
            self.create_configmap()
            self.log.debug(
                "Creating workflow in namespace {}: '{}'".format(
                    namespace, workflow))
            wf = api.create_namespaced_workflow(namespace, workflow)
            return wf

    def create_configmap(self):
        with start_action(action_type="create_configmap"):
            api = self.parent.api  # Core, not Workflow
            namespace = self.parent.namespace_mgr.namespace
            cfgmap = self.cfg_map
            try:
                self.log.info(
                    "Attempting to create configmap in {}".format(namespace))
                api.create_namespaced_config_map(namespace, cfgmap)
            except ApiException as e:
                if e.status != 409:
                    estr = "Create configmap failed: {}".format(e)
                    self.log.exception(estr)
                    raise
                else:
                    self.log.info("Configmap already exists.")

    def submit_workflow(self, data):
        with start_action(action_type="submit_workflow"):
            self.define_workflow(data)
            nm = self.parent.namespace_mgr
            nm.ensure_namespace()
            nm.ensure_namespaced_service_account()
            wf = self.create_workflow()
            return wf

    def delete_workflow(self, wfid):
        with start_action(action_type="delete_workflow"):
            namespace = self.parent.namespace_mgr.namespace
            api = self.parent.wf_api
            wf = api.delete_namespaced_workflow(namespace, wfid)
            return wf

    def get_workflow(self, wfid):
        with start_action(action_type="get_workflow"):
            namespace = self.parent.namespace_mgr.namespace
            api = self.parent.wf_api
            wf = api.get_namespaced_workflow(namespace, wfid)
            return wf

    def dump(self):
        '''Return contents dict for aggregation and pretty-printing.
        '''
        wd = {"parent": str(self.parent),
              "cmd_vol": str(self.cmd_vol),
              "workflow": str(self.workflow),
              "cfg_map": self.cfg_map,
              "wf_input": self.wf_input
              }
        return wd

    def toJSON(self):
        return json.dumps(self.dump())


class LSSTWorkflow(Workflow):
    parms = {}
    entrypoint = "noninteractive"
    run_as_user = 769
    run_as_group = 769

    def __init__(self, *args, **kwargs):
        self.parms = kwargs.pop('parms')
        super().__init__(*args, **kwargs)
        self.spec.volumes = self.parms['vols']
        lbl = {'argocd.argoproj.io/instance': 'nublado-users'}
        self.metadata.labels = lbl
        self.metadata.generate_name = self.parms['name'] + '-'
        self.metadata.name = None
        username = self.parms['username']
        account = "{}-svcacct".format(username)
        self.spec.service_account_name = account
        self.metadata.annotations = self.build_annotations()

    def build_annotations(self):
        '''Add annotations telling Argo CD to not prune these resources or to
        count them against the sync state.
        '''
        with start_action(action_type="build_annotations"):
            anno = {}
            anno['argocd.argoproj.io/compare-options'] = 'IgnoreExtraneous'
            anno['argocd.argoproj.io/sync-options'] = 'Prune=false'

            # Enable multus if requested
            if self.parms['enable_multus']:
                ks = 'kube-system/macvlan-conf'
                # This probably ought to be a config parameter
                anno['k8s.v1.cni.cncf.io/networks'] = ks

            cmdlist = self.parms['command']
            if not cmdlist:
                return anno
            annobase = "lsst.org/wf_cmd"
            idx = 0
            for cmd in cmdlist:
                akey = annobase + "_{}".format(idx)
                if len(cmd) < 64:
                    anno[akey] = cmd
                else:
                    jdx = 0
                    while cmd:
                        chunk = cmd[:63]
                        cmd = cmd[63:]
                        skey = akey + "_{}".format(jdx)
                        anno[skey] = chunk
                        jdx = jdx+1
                    idx = idx + 1
            return anno

    @template
    def noninteractive(self) -> V1Container:
        with start_action(action_type="noninteractive_container"):
            container = V1Container(
                command=["/opt/lsst/software/jupyterlab/provisionator.bash"],
                args=[],
                image=self.parms["image"],
                name=self.parms["name"],
                env=self.parms["env"],
                image_pull_policy="Always",
                volume_mounts=self.parms["vmts"],
                resources=V1ResourceRequirements(
                    limits={"memory": self.parms['mem_limit'],
                            "cpu": self.parms['cpu_limit']},
                    requests={"memory": self.parms['mem_guar'],
                              "cpu": self.parms['cpu_guar']}
                ),
                security_context=V1SecurityContext(
                    run_as_group=self.run_as_group,
                    run_as_user=self.run_as_user,
                )
            )
            self.volumes = self.parms['vols']
            lbl = {'argocd.argoproj.io/instance': 'nublado-users'}
            self.metadata.labels = lbl
            self.metadata.generate_name = self.parms['name'] + '-'
            self.metadata.name = None
            self.service_account_name = self.parms['username'] + '-svcacct'

            return container



import datetime
import os
import requests
from eliot import start_action
from . import SingletonScanner


class Reaper(SingletonScanner):
    '''Class to allow implementation of image retention policy.
    '''

    _categorized_tags = {"weekly": [],
                         "daily": [],
                         "experimental": []
                         }
    # We don't need to categorize releases since we never delete any of
    #  them.

    def __init__(self, *args, **kwargs):
        self.keep_experimentals = kwargs.pop('keep_experimentals', 10)
        self.keep_dailies = kwargs.pop('keep_dailies', 15)
        self.keep_weeklies = kwargs.pop('keep_weeklies', 78)
        self.dry_run = kwargs.pop('dry_run', False)
        self.more_cowbell = self.reap
        super().__init__(**kwargs)
        self.logger.debug(("Keeping: {} weeklies, {} dailies, and {} " +
                           "experimentals.").format(self.keep_weeklies,
                                                    self.keep_dailies,
                                                    self.keep_experimentals))
        self.delete_tags = False
        if self.registry_url.startswith('registry.hub.docker.com'):
            self.delete_tags = True
        self.reapable = {}

    def _categorize_tags(self):
        with start_action(action_type="_categorize_tags"):
            tags = self.get_all_tags()  # Should wait for initial scan
            for t in tags:
                if t.startswith('w'):
                    self._categorized_tags["weekly"].append(t)
                elif t.startswith('d'):
                    self._categorized_tags["daily"].append(t)
                elif t.startswith('exp'):
                    self._categorized_tags["experimental"].append(t)
            for i in ["experimental", "daily", "weekly"]:
                self._categorized_tags[i].sort(
                    key=lambda tag: datetime.datetime.strptime(
                        self._results_map[tag]["last_updated"].replace(
                            "Z", "UTC"),
                        "%Y-%m-%dT%H:%M:%S.%f%Z"))

    def _select_victims(self):
        with start_action(action_type="_select victims"):
            self._categorize_tags()
            reaptags = []
            sc = self._categorized_tags
            reaptags.extend(sc["experimental"][:-(self.keep_experimentals)])
            reaptags.extend(sc["daily"][:-(self.keep_dailies)])
            reaptags.extend(sc["weekly"][:-(self.keep_weeklies)])
            reapable = {}
            for r in reaptags:
                reapable[r] = self._results_map[r]["hash"]
            self.logger.debug("Images to reap: {}.".format(reapable))
            self.reapable = reapable

    def report_reapable(self):
        '''Return a space-separated list of reapable images.
        '''
        with start_action(action_type="report_reapable"):
            self._select_victims()
            return " ".join(self.reapable.keys())

    def reap(self):
        '''Select and delete images.
        '''
        with start_action(action_type="reap"):
            self._select_victims()
            self._delete_from_repo()

    def _delete_from_repo(self):
        with start_action(action_type="_delete_from_repo"):
            tags = list(self.reapable.keys())
            if not tags:
                self.logger.info("No images to reap.")
                return
            if self.dry_run:
                self.logger.info("Dry run: images to reap: {}".format(tags))
                return
            headers = {
                "Accept": ("application/vnd.docker.distribution.manifest." +
                           "v2+json")}
            sc = 0
            if self.registry_url.startswith("https://registry.hub.docker.com"):
                self._delete_tags_from_docker_hub()
                return
            for t in tags:
                self.logger.debug("Attempting to reap '{}'.".format(t))
                h = self.reapable[t]
                path = self.registry_url + "manifests/" + h
                resp = requests.delete(path, headers=headers)
                sc = resp.status_code
                if sc == 401:
                    auth_hdr = self._authenticate_to_repo(resp)
                    headers.update(auth_hdr)  # Retry with new auth
                    self.logger.warning("Retrying with new authentication.")
                    resp = requests.delete(path, headers=headers)
                    sc = resp.status_code
                if (sc >= 200) and (sc < 300):
                    # Got it.
                    del(self._results_map[t])
                else:
                    self.logger.warning("DELETE {} => {}".format(path, sc))
                    self.logger.warning("Headers: {}".format(resp.headers))
                    self.logger.warning("Body: {}".format(resp.text))
                if self.cachefile:
                    self._writecachefile()  # Remove deleted tags

    def _delete_tags_from_docker_hub(self):
        # This is, of course, completely different from the published API
        #  https://github.com/docker/hub-feedback/issues/496
        with start_action(action_type="_delete_tags_from_docker_hub"):
            self.logger.info("Deleting tags from Docker Hub.")
            r_user = os.getenv("IMAGE_REAPER_USER")
            r_pw = os.getenv("IMAGE_REAPER_PASSWORD")
            data = {"username": r_user,
                    "password": r_pw}
            headers = {"Content-Type": "application/json",
                       "Accept": "application/json"}
            token = None
            # Exchange username/pw for token
            if r_user and r_pw:
                resp = requests.post("https://hub.docker.com/v2/users/login",
                                     headers=headers, json=data)
                r_json = resp.json()
                if r_json:
                    token = r_json.get("token")
                else:
                    self.logger.warning("Failed to authenticate:")
                    self.logger.warning("Headers: {}".format(resp.headers))
                    self.logger.warning("Body: {}".format(resp.text))
            else:
                self.logger.error("Did not have username and password.")
            if not token:
                self.logger.error("Could not acquire JWT token.")
                return
            headers["Authorization"] = "JWT {}".format(token)
            tags = list(self.reapable.keys())
            for t in tags:
                path = ("https://hub.docker.com/v2/repositories/" +
                        self.owner + "/" + self.name + "/tags/" + t + "/")
                self.logger.info("Deleting tag '{}'".format(t))
                resp = requests.delete(path, headers=headers)
                sc = resp.status_code
                if (sc < 200) or (sc >= 300):
                    self.logger.warning("DELETE {} => {}".format(path, sc))
                    self.logger.warning("Headers: {}".format(resp.headers))
                    self.logger.warning("Body: {}".format(resp.text))
                    if sc != 404:
                        continue
                    # It's already gone, so remove from map!
                del(self._results_map[t])
            if self.cachefile:
                self._writecachefile()  # Remove deleted tags

    def _authenticate_to_repo(self, resp):
        with start_action(action_type="_authenticate_to_repo"):
            self.logger.warning("Authentication Required.")
            self.logger.warning("Headers: {}".format(resp.headers))
            self.logger.warning("Body: {}".format(resp.text))
            magicheader = resp.headers['Www-Authenticate']
            if magicheader[:7] == "Bearer ":
                hd = {}
                hl = magicheader[7:].split(",")
                for hn in hl:
                    il = hn.split("=")
                    kk = il[0]
                    vv = il[1].replace('"', "")
                    hd[kk] = vv
                if (not hd or "realm" not in hd or "service" not in hd
                        or "scope" not in hd):
                    return None
                endpoint = hd["realm"]
                del hd["realm"]
                # We need to glue in authentication for DELETE, and that alas
                #  means a userid and password.
                r_user = os.getenv("IMAGE_REAPER_USER")
                r_pw = os.getenv("IMAGE_REAPER_PASSWORD")
                auth = None
                if r_user and r_pw:
                    auth = (r_user, r_pw)
                    self.logger.warning("Added Basic Auth credentials")
                headers = {
                    "Accept": ("application/vnd.docker.distribution." +
                               "manifest.v2+json")
                }
                self.logger.warning(
                    "Requesting auth scope {}".format(hd["scope"]))
                tresp = requests.get(endpoint, headers=headers, params=hd,
                                     json=True, auth=auth)
                jresp = tresp.json()
                authtok = jresp.get("token")
                if authtok:
                    self.logger.info("Received an auth token.")
                    self.logger.warning("{}".format(authtok))
                    return {"Authorization": "Bearer {}".format(authtok)}
                else:
                    self.logger.error("No auth token: {}".format(jresp))
            return {}



import json
import logging
import os
import sys
from eliot import to_file, start_action
from eliot.stdlib import EliotHandler
from jupyter_client.localinterfaces import public_ips
from urllib.parse import urlparse
from .. import Singleton
from ..utils import (str_bool, listify, intify, floatify, make_logger,
                     get_execution_namespace, sanitize_dict)


class LSSTConfig(metaclass=Singleton):
    '''LSSTConfig is a Singleton bag of attributes to hold
    important-to-the-LSST JupyterHub environment variables and default
    values.
    '''

    def __init__(self, *args, **kwargs):
        default_level = logging.WARNING
        # we can't check this prior to the config existing...
        debug = str_bool(os.getenv('DEBUG', False))
        if debug:
            default_level = logging.DEBUG
        self._setup_logger(level=default_level)
        self.source = kwargs.pop('source', 'environment')
        if self.source != 'environment':
            raise ValueError(
                "'environment' is currently only allowed config source!")
        # To make it easy to read settings from a document later on...
        if self.source == 'environment':
            self.load_from_environment()
        if self.debug:
            # Reset default loglevel
            rootlogger = logging.getLogger()
            level = rootlogger.getEffectiveLevel()
            if level > logging.DEBUG:
                rootlogger.warn("Setting root logging level to DEBUG.")
                rootlogger.setLevel(logging.DEBUG)
        self.create_derived_settings()

    def _setup_logger(self, level=logging.WARNING):
        # In Python 3.8 we could use force=True to force the root logger
        #  to override existing defaults
        # Add our formats
        with start_action(action_type="_setup_logger"):
            fstr = '[%(levelname).1s %(asctime)s.%(msecs).03d'
            fstr += ' %(module)s:%(funcName)s:%(lineno)d] %(message)s'
            dstr = '%Y-%m-%d %H:%M:%S'
            self.log_format = fstr
            self.log_datefmt = dstr
            self.log_level = level
            rootlogger = logging.getLogger()
            rootlogger.handlers = [EliotHandler()]
            to_file(sys.stderr)
            loggers = []
            lrmld = logging.root.manager.loggerDict
            loggers = loggers + [logging.getLogger(name) for name in lrmld]
            rootlogger.debug("Removing all existing log handlers.")
            for lgr in loggers:
                rootlogger.debug(
                    "Removing log handlers for '{}'".format(lgr.name))
                lgr.handlers = []
            rootlogger.debug("Setting default log level to '{}'".format(level))
            rootlogger.setLevel(level)
            # There are some things we just don't want to log at DEBUG
            chattycathies = {'kubernetes': logging.WARNING,
                             'urllib3': logging.WARNING,
                             'JupyterHub': logging.INFO}
            for k in chattycathies:
                v = chattycathies[k]
                lgr = logging.getLogger(k)
                el = lgr.getEffectiveLevel()
                if el < v:
                    rootlogger.debug(
                        "Logger {} @level {} -> {}.".format(k, el, v))
                    lgr.setLevel(v)
            self.log = make_logger(level=level)

    def load_from_environment(self):
        '''Populate attributes from environment variables.
        '''
        with start_action(action_type="load_from_environment"):
            self.debug = str_bool(os.getenv('DEBUG'))
            # Authentication parameters
            self.authenticator_type = (os.getenv('AUTH_PROVIDER') or
                                       os.getenv('OAUTH_PROVIDER') or
                                       'github')
            self.oauth_client_id = os.getenv('OAUTH_CLIENT_ID')
            self.oauth_client_secret = os.getenv('OAUTH_CLIENT_SECRET')
            self.oauth_callback_url = os.getenv('OAUTH_CALLBACK_URL')
            # Authenticator-specific parameters
            self.jwt_logout_url = os.getenv("LOGOUT_URL") or '/logout'
            self.jwt_signing_certificate = (
                os.getenv('JWT_SIGNING_CERTIFICATE') or
                '/opt/jwt/signing-certificate.pem')
            self.cilogon_host = os.getenv('CILOGON_HOST') or 'cilogon.org'
            self.cilogon_skin = os.getenv('CILOGON_SKIN') or 'LSST'
            self.cilogon_idp = os.getenv('CILOGON_IDP_SELECTION')
            self.cilogon_allowlist = listify(
                os.getenv('CILOGON_GROUP_WHITELIST'))
            self.cilogon_denylist = listify(
                os.getenv('CILOGON_GROUP_DENYLIST'))
            self.strict_ldap_groups = str_bool(os.getenv('STRICT_LDAP_GROUPS'))
            self.github_host = os.getenv('GITHUB_HOST') or 'github.com'
            self.github_allowlist = listify(
                os.getenv('GITHUB_ORGANIZATION_WHITELIST'))
            self.github_denylist = listify(
                os.getenv('GITHUB_ORGANIZATION_DENYLIST'))
            # Settings for Options Form
            self.form_selector_title = (os.getenv('LAB_SELECTOR_TITLE') or
                                        'Container Image Selector')
            self.form_template = (
                os.getenv('OPTIONS_FORM_TEMPLATE') or
                ('/opt/lsst/software/jupyterhub/templates/' +
                 'options_form.template.html'))
            self.form_sizelist = (listify(
                os.getenv('OPTIONS_FORM_SIZELIST')) or
                ['tiny', 'small', 'medium', 'large'])
            self.max_scan_delay = intify(os.getenv('MAX_SCAN_DELAY'), 300)
            self.initial_scan_interval = floatify(
                os.getenv('INITIAL_SCAN_INTERVAL'), 0.2)
            self.max_scan_interval = floatify(
                os.getenv('MAX_SCAN_INTERVAL'), 5.0)
            self.tiny_cpu_max = floatify(os.getenv('TINY_CPU_MAX'), 0.5)
            self.mb_per_cpu = intify(os.getenv('MB_PER_CPU'), 2048)
            self.size_index = intify(os.getenv('SIZE_INDEX'), 1)
            # Settings for Quota Manager
            self.resource_map = (os.getenv('RESOURCE_MAP') or
                                 '/opt/lsst/software/jupyterhub/resources/' +
                                 'resourcemap.json')
            self.max_dask_workers = int(os.getenv('MAX_DASK_WORKERS', 25))
            # Settings for Volume Manager
            self.volume_definition_file = (
                os.getenv('VOLUME_DEFINITION_FILE') or
                ('/opt/lsst/software/jupyterhub/' +
                 'mounts/mountpoints.json'))
            # Hub settings for Lab spawning
            self.lab_uid = intify(os.getenv('LAB_UID'), 769)
            self.lab_gid = intify(os.getenv('LAB_GID'), 769)
            self.lab_fs_gid = intify(os.getenv('LAB_FS_GID'), self.lab_gid)
            self.lab_default_image = (os.getenv('LAB_IMAGE') or
                                      "lsstsqre/sciplat-lab:latest")
            self.lab_size_range = os.getenv('LAB_SIZE_RANGE') or '4.0'
            self.cull_timeout = os.getenv('LAB_CULL_TIMEOUT') or '64800'
            self.cull_policy = os.getenv('LAB_CULL_POLICY') or 'idle:remote'
            self.allow_dask_spawn = str_bool(os.getenv('ALLOW_DASK_SPAWN'))
            self.restrict_dask_nodes = os.getenv('RESTRICT_DASK_NODES')
            self.restrict_lab_nodes = os.getenv('RESTRICT_LAB_NODES')
            self.lab_nodejs_max_mem = os.getenv('LAB_NODEJS_MAX_MEM') or '6144'
            self.external_hub_url = os.getenv('EXTERNAL_HUB_URL')
            self.hub_route = os.getenv('HUB_ROUTE') or '/'
            self.external_instance_url = os.getenv('EXTERNAL_INSTANCE_URL')
            self.firefly_route = os.getenv('FIREFLY_ROUTE') or '/firefly'
            self.js9_route = os.getenv('JS9_ROUTE') or '/js9'
            self.api_route = os.getenv('API_ROUTE') or '/api'
            self.tap_route = os.getenv('TAP_ROUTE') or '/api/tap'
            self.soda_route = os.getenv('SODA_ROUTE') or '/api/image/soda'
            self.workflow_route = os.getenv('WORKFLOW_ROUTE') or '/wf'
            self.external_firefly_url = os.getenv('EXTERNAL_FIREFLY_URL')
            self.external_js9_url = os.getenv('EXTERNAL_JS9_URL')
            self.external_api_url = os.getenv('EXTERNAL_API_URL')
            self.external_tap_url = os.getenv('EXTERNAL_TAP_URL')
            self.external_soda_url = os.getenv('EXTERNAL_SODA_URL')
            self.external_workflow_url = os.getenv('EXTERNAL_WORKFLOW_URL')
            self.auto_repo_urls = os.getenv('AUTO_REPO_URLS')
            # Prepuller settings
            self.lab_repo_owner = os.getenv('LAB_REPO_OWNER') or 'lsstsqre'
            self.lab_repo_name = os.getenv('LAB_REPO_NAME') or 'sciplat-lab'
            self.lab_repo_host = os.getenv('LAB_REPO_HOST') or 'hub.docker.com'
            self.prepuller_namespace = (os.getenv('PREPULLER_NAMESPACE') or
                                        get_execution_namespace())
            self.prepuller_experimentals = intify(
                os.getenv('PREPULLER_EXPERIMENTALS'), 0)
            self.prepuller_dailies = intify(os.getenv('PREPULLER_DAILIES'), 3)
            self.prepuller_weeklies = intify(
                os.getenv('PREPULLER_WEEKLIES'), 2)
            self.prepuller_releases = intify(
                os.getenv('PREPULLER_RELEASES'), 1)
            self.prepuller_cachefile = os.getenv('PREPULLER_CACHEFILE',
                                                 '/tmp/repo-cache.json')
            self.prepuller_timeout = os.getenv('PREPULLER_TIMEOUT', 3300)
            cstr = "echo \"Prepuller for $(hostname) completed at $(date).\""
            self.prepuller_command = ["/bin/sh", "-c", cstr]
            prp_cmd = os.getenv('PREPULLER_COMMAND_JSON')
            if prp_cmd:
                self.prepuller_command = json.loads(prp_cmd)
            # Reaper settings
            self.reaper_keep_experimentals = intify(
                os.getenv('REAPER_KEEP_EXPERIMENTALS'), 10)
            self.reaper_keep_dailies = intify(
                os.getenv('REAPER_KEEP_DAILIES'), 15)
            self.reaper_keep_weeklies = intify(
                os.getenv('REAPER_KEEP_WEEKLIES'), 78)
            # Fileserver settings
            self.fileserver_host = (os.getenv('EXTERNAL_FILESERVER_IP') or
                                    os.getenv('FILESERVER_SERVICE_HOST'))
            # Reaper settings
            self.reaper_user = os.getenv('IMAGE_REAPER_USER')
            self.reaper_password = os.getenv('IMAGE_REAPER_PASSWORD')
            # Hub internal settings
            my_ip = public_ips()[0]
            self.helm_tag = os.getenv('HELM_TAG')
            self.hub_host = os.getenv('HUB_SERVICE_HOST') or my_ip
            self.hub_api_port = os.getenv('HUB_SERVICE_PORT_API') or 8081
            self.proxy_host = os.getenv('PROXY_SERVICE_HOST') or my_ip
            self.proxy_api_port = os.getenv('PROXY_SERVICE_PORT_API') or 8001
            self.session_db_url = os.getenv('SESSION_DB_URL')
            # Pod multicast (used with SAL)
            self.enable_multus = str_bool(os.getenv('ENABLE_MULTUS'))
            # DDS domain (used with SAL)
            self.lab_dds_domain = os.getenv('LSST_DDS_DOMAIN')
            # These have to be set post-initialization to avoid a circular
            #  dependency.
            self.authenticator_class = None
            self.spawner_class = None

    def create_derived_settings(self):
        '''Create further settings from passed-in ones.
        '''
        with start_action(action_type="create_derived_settings"):
            self.proxy_api_url = 'http://{}:{}'.format(
                self.proxy_host, self.proxy_api_port)
            if self.github_host == 'github.com':
                self.github_api = 'api.github.com'
            else:
                self.github_api = "{}/api/v3".format(self.github_host)
            audience = None
            callback_url = self.oauth_callback_url
            if callback_url:
                netloc = urlparse(callback_url).netloc
                scheme = urlparse(callback_url).scheme
                if netloc and scheme:
                    audience = scheme + "://" + netloc
            if not audience:
                audience = self.oauth_client_id or ''
            self.audience = audience
            self.bind_url = 'http://0.0.0.0:8000{}'.format(self.hub_route)
            self.hub_bind_url = 'http://0.0.0.0:8081'.format(self.hub_route)
            self.hub_connect_url = 'http://{}:{}{}'.format(self.hub_host,
                                                           self.hub_api_port,
                                                           self.hub_route)
            mm = self.lab_nodejs_max_mem
            self.lab_node_options = None
            if mm:
                self.lab_node_options = "--max-old-space-size={}".format(mm)
            while self.hub_route.endswith('/') and self.hub_route != '/':
                self.hub_route = self.hub_route[:-1]
            if not self.external_hub_url:
                oauth_callback = self.oauth_callback_url
                endstr = '/hub/oauth_callback'
                if oauth_callback and oauth_callback.endswith(endstr):
                    self.external_hub_url = oauth_callback[:-len(endstr)]
            if not self.external_instance_url:
                ehu = self.external_hub_url
                if ehu:
                    if ehu.endswith(self.hub_route):
                        self.external_instance_url = ehu[:-len(self.hub_route)]

    def dump(self):
        '''Return dict for pretty printing.
        '''
        myvars = vars(self)
        sanitized = sanitize_dict(myvars, ['oauth_client_secret',
                                           'reaper_password',
                                           'session_db_url'])
        # Stringify classrefs
        for key in ["log", "authenticator_class", "spawner_class"]:
            val = sanitized.get(key)
            if val:
                sanitized[key] = str(val)
        return sanitized

    def toJSON(self):
        return self.dump()



'''Setup module for LSST JupyterHub Utilities.
'''
import codecs
import io
import os
import setuptools


def get_version(file, name='__version__'):
    '''Get the version of the package from the given file by
    executing it and extracting the given `name`.
    '''
    path = os.path.realpath(file)
    version_ns = {}
    with io.open(path, encoding="utf8") as f:
        exec(f.read(), {}, version_ns)
    return version_ns[name]


def local_read(filename):
    '''Convenience function for includes.
    '''
    full_filename = os.path.join(
        os.path.abspath(os.path.dirname(__file__)),
        filename)
    return codecs.open(full_filename, 'r', 'utf-8').read()


NAME = 'jupyterhubutils'
DESCRIPTION = 'LSST utilites for Jupyter Hub server environment'
LONG_DESCRIPTION = local_read("README.md")
VERSION = get_version('%s/_version.py' % NAME)
AUTHOR = 'Adam Thornton'
AUTHOR_EMAIL = 'athornton@lsst.org'
URL = 'https://github.com/sqre-lsst/jupyterhubutils'
LICENSE = 'MIT'

setuptools.setup(
    name=NAME,
    version=get_version("%s/_version.py" % NAME),
    long_description=LONG_DESCRIPTION,
    packages=setuptools.find_packages(),
    url=URL,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    license=LICENSE,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'License :: OSI Approved :: MIT License',
    ],
    keywords='lsst',
    install_requires=[
        'requests>=2.0.0,<3.0.0',
        'kubernetes>=10.0.0,<11.0.0',
        'semver>=2.0.0,<3.0.0',
        'oauthenticator>=0.9.0,<1.0.0',
        'jupyter-client>=5.0.0,<7.0.0',
        'jupyterhub-jwtauthenticator>=0.1.0,<1.0.0',
        'jupyterhub-kubespawner>=0.10.0,<1.0.00',
        'jinja2>=2.0.0,<3.0.0',
        'pytz>=2019.3',
        'eliot',
        'eliot-tree',
        'argo-workflows>=3.0.0,<3.2.0',
        'argo-workflows-sdk',
        'asgiref>=3.0.0,<4.0.0'
    ],
    entry_points={
        'console_scripts': [
            'prepuller = jupyterhubutils.scanrepo:prepullerstandalone',
            'reaper = jupyterhubutils.scanrepo:reaperstandalone',
            'scanrepo = jupyterhubutils.scanrepo:standalone'
        ],
        'jupyterhub.authenticators': [
            ('LSSTGitHubOAuthenticator = ' +
             'jupyterhubutils.authenticator.LSSTGitHubOAuthenticator'),
            ('LSSTCILogonOAuthenticator = ' +
             'jupyterhubutils.authenticator.LSSTCILogonOAuthenticator'),
            ('LSSTJWTAuthenticator = ' +
             'jupyterhubutils.authenticator.LSSTJWTAuthenticator'),
        ],
        'jupyterhub.spawners': [
            'LSSTSpawner = jupyterhubutils.spawners.LSSTSpawner',
        ],
    }
)



import base64
import json
import os
import yaml
from eliot import start_action
from kubernetes import client
from .. import LoggableChild


class LSSTVolumeManager(LoggableChild):
    '''Class to provide support for document-driven Volume assignment.
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.volume_list = []
        self.k8s_volumes = []
        self.k8s_vol_mts = []

    def make_volumes_from_config(self):
        '''Create volume definition representation from document.
        Override this in a subclass if you like.
        '''
        with start_action(action_type="make_volumes_from_config"):
            vollist = []
            config = []
            config_file = self.parent.config.volume_definition_file
            if not os.path.exists(config_file):
                return vollist
            with open(config_file, "r") as fp:
                config = json.load(fp)
            for mtpt in config:
                mountpoint = mtpt["mountpoint"]  # Fatal error if it
                # doesn't exist
                if mtpt.get("disabled"):
                    continue
                if mountpoint[0] != "/":
                    mountpoint = "/" + mountpoint
                host = (mtpt.get("fileserver-host") or
                        self.parent.config.fileserver_host)
                export = mtpt.get("fileserver-export") or (
                    "/exports" + mountpoint)
                mode = (mtpt.get("mode") or "ro").lower()
                k8s_vol = mtpt.get("kubernetes-volume")
                if k8s_vol:
                    raise ValueError("Shadow PVs and matching PVCs " +
                                     "are no longer supported!")
                hostpath = mtpt.get("hostpath")
                vollist.append({
                    "mountpoint": mountpoint,
                    "hostpath": hostpath,
                    "host": host,
                    "export": export,
                    "mode": mode,
                })
            self.volume_list = vollist
            self.log.debug("Volumes: {}".format(vollist))
            self._define_k8s_object_representations()

    def _define_k8s_object_representations(self):
        with start_action(action_type="_define_k8s_object_representations"):
            self.k8s_volumes = []
            self.k8s_vol_mts = []
            for vol in self.volume_list:
                k8svol = None
                k8smt = None
                if vol.get("hostpath"):
                    k8svol = self._define_k8s_hostpath_vol(vol)
                else:
                    k8svol = self._define_k8s_nfs_vol(vol)
                    k8smt = self._define_k8s_mtpt(vol)
                if k8svol and k8smt:
                    self.k8s_volumes.append(k8svol)
                    self.k8s_vol_mts.append(k8smt)

    def _define_k8s_hostpath_vol(self, vol):
        with start_action(action_type="_define_k8s_hostpath_vol"):
            return client.V1Volume(
                name=self._get_volume_name_for_mountpoint(vol["mountpoint"]),
                host_path=client.V1HostPath(
                    path=vol["mountpoint"]
                )
            )

    def _define_k8s_nfs_vol(self, vol):
        with start_action(action_type="_define_k8s_nfs_vol"):
            knf = client.V1NFSVolumeSource(
                path=vol["export"],
                server=vol["host"]
            )
            if vol["mode"] == "ro":
                knf.read_only = True
            return client.V1Volume(
                name=self._get_volume_name_for_mountpoint(vol["mountpoint"]),
                nfs=knf
            )

    def _define_k8s_mtpt(self, vol):
        with start_action(action_type="_define_k8s_mtpt"):
            mt = client.V1VolumeMount(
                mount_path=vol["mountpoint"],
                name=self._get_volume_name_for_mountpoint(vol["mountpoint"]),
            )
            if vol["mode"] == "ro":
                mt.read_only = True
            return mt

    def _get_volume_name_for_mountpoint(self, mountpoint):
        with start_action(action_type="_get_volume_name_for_mountpoint"):
            return mountpoint[1:].replace('/', '-')

    def _get_volume_yaml_str(self, left_pad=0):
        with start_action(action_type="_get_volume_yaml_str"):
            vols = self.k8s_volumes
            if not vols:
                self.log.warning("No volumes defined.")
                return ''
            vl = []
            for vol in vols:
                nm = vol.name
                hp = vol.host_path
                nf = vol.nfs
                vo = {"name": nm}
                if hp:
                    vo["hostPath"] = {"path": hp.path}
                elif nf:
                    am = "ReadWriteMany"
                    if nf.read_only:
                        am = "ReadOnlyMany"
                    vo["nfs"] = {"server": nf.server,
                                 "path": nf.path,
                                 "accessMode": am}
                    vl.append(vo)
            vs = {"volumes": vl}
            ystr = yaml.dump(vs)
            return self._left_pad(ystr, left_pad)

    def _get_volume_mount_yaml_str(self, left_pad=0):
        with start_action(action_type="_get_volume_mount_yaml_str"):
            vms = self.k8s_vol_mts
            if not vms:
                self.log.warning("No volume mounts defined.")
                return ''
            vl = []
            for vm in vms:
                vo = {}
                vo["name"] = vm.name
                vo["mountPath"] = vm.mount_path
                if vm.read_only:
                    vo["readOnly"] = True
                vl.append(vo)
            vs = {"volumeMounts": vl}
            ystr = yaml.dump(vs)
            return self._left_pad(ystr, left_pad)

    def _left_pad(self, line_str, left_pad=0):
        with start_action(action_type="_left_pad"):
            pad = ' ' * left_pad
            ylines = line_str.split("\n")
            padlines = [pad + l for l in ylines]
            return "\n".join(padlines)

    def get_dask_volume_b64(self):
        '''Return the base-64 encoding of the K8s statements to create
        the pod's mountpoints.  Probably better handled as a ConfigMap.
        '''
        with start_action(action_type="get_dask_volume_b64"):
            vmt_yaml_str = self._get_volume_mount_yaml_str(left_pad=4)
            vol_yaml_str = self._get_volume_yaml_str(left_pad=2)
            ystr = "{}\n{}".format(vmt_yaml_str, vol_yaml_str)
            benc = base64.b64encode(ystr.encode('utf-8')).decode('utf-8')
            return benc

    def dump(self):
        '''Return contents dict for aggregation and pretty-printing.
        '''
        vd = {"parent": str(self.parent),
              "volume_list": self.volume_list,
              "k8s_volumes": [str(x) for x in self.k8s_volumes],
              "k8s_vol_mts": [str(x) for x in self.k8s_vol_mts]}
        return vd

    def toJSON(self):
        return json.dumps(self.dump())
